/*
 *  OMIKUJI Machine for STM32F401re nucleo board
 *
 *  Copyright(C) by YI Studio
 */

/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2004-2012 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 * 
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 * 
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 * 
 *  $Id: sample1.c 2728 2015-12-30 01:46:11Z ertl-honda $
 */

#include <string.h>
#include <stdlib.h>

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "kernel_cfg.h"
#include "omikuji.h"

#include "ff.h"
#include "diskio.h"
#include "spi.h"

#include "dev_uart.h"
#include "printout.h"
#include "dev_access.h"


#define DEF_FONT_SIZE          16
#define DEF_FONT_A16_VAR       jis8x16rk_fnt
#define DEF_FONT_K16_VAR       jiskan16_fnt

#define DEF_NUM_FILE           "num.txt"
#define DEF_LEN_FILE           "len.txt"
#define DEF_PRB_FILE           "prob.txt"

#define DEF_NUM_INIT           1
#define DEF_NUM_MAXCOL         2
#define DEF_NUM_MAX            20
#define DEF_LEN_INIT           96
#define DEF_LEN_MAXCOL         3
#define DEF_LEN_MAX            129
#define DEF_READBUFF           0x1000

int  var_debug_mode ;
int  var_data_num ;
int  var_data_len ;
int  var_data_prb_total ;
int  var_data_prb[ DEF_NUM_MAX ];
int  var_data_prs[ DEF_NUM_MAX ];
int  var_data_dst[ DEF_NUM_MAX ];
BYTE  readbuff[ DEF_READBUFF ];

FATFS  var_fs;
FATFS  var_fs2;

extern  unsigned long  SystemCoreClock ;


/*
 *  サービスコールのエラーのログ出力
 */
Inline void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
	if (ercd < 0) {
		t_perror(LOG_ERROR, file, line, expr, ercd);
	}
}

#define	SVC_PERROR(expr)	svc_perror(__FILE__, __LINE__, #expr, (expr))


/*
 *  ループ回数
 */
ulong_t	tex_loop;		/* 例外処理ルーチン内でのループ回数 */


/*
 *  並行して実行されるタスク用のタスク例外処理ルーチン
 */
void tex_routine(TEXPTN texptn, intptr_t exinf)
{
	volatile ulong_t	i;
	int_t	tskno = (int_t) exinf;

	syslog(LOG_NOTICE, "task%d receives exception 0x%04x.", tskno, texptn);
	for (i = 0; i < tex_loop; i++);

	if ((texptn & 0x8000U) != 0U) {
		syslog(LOG_INFO, "#%d#ext_tsk()", tskno);
		SVC_PERROR(ext_tsk());
		assert(0);
	}
}

/*
 *  CPU例外ハンドラ
 */
#ifdef CPUEXC1

void
cpuexc_handler(void *p_excinf)
{
	ID		tskid;

	syslog(LOG_NOTICE, "CPU exception handler (p_excinf = %08p).", p_excinf);
	if (sns_ctx() != true) {
		syslog(LOG_WARNING,
					"sns_ctx() is not true in CPU exception handler.");
	}
	if (sns_dpn() != true) {
		syslog(LOG_WARNING,
					"sns_dpn() is not true in CPU exception handler.");
	}
	syslog(LOG_INFO, "sns_loc = %d sns_dsp = %d sns_tex = %d",
									sns_loc(), sns_dsp(), sns_tex());
	syslog(LOG_INFO, "xsns_dpn = %d xsns_xpn = %d",
									xsns_dpn(p_excinf), xsns_xpn(p_excinf));

	if (xsns_xpn(p_excinf)) {
		syslog(LOG_NOTICE, "Sample program ends with exception.");
		SVC_PERROR(ext_ker());
		assert(0);
	}

	SVC_PERROR(iget_tid(&tskid));
	SVC_PERROR(iras_tex(tskid, 0x8000U));
}

#endif /* CPUEXC1 */

/*
 *  周期ハンドラ
 *
 *  HIGH_PRIORITY，MID_PRIORITY，LOW_PRIORITY の各優先度のレディキュー
 *  を回転させる．
 */
void cyclic_handler(intptr_t exinf)
{
	SVC_PERROR(irot_rdq(HIGH_PRIORITY));
	SVC_PERROR(irot_rdq(MID_PRIORITY));
	SVC_PERROR(irot_rdq(LOW_PRIORITY));
}

/*
 *  アラームハンドラ
 *
 *  HIGH_PRIORITY，MID_PRIORITY，LOW_PRIORITY の各優先度のレディキュー
 *  を回転させる．
 */
void alarm_handler(intptr_t exinf)
{
	SVC_PERROR(irot_rdq(HIGH_PRIORITY));
	SVC_PERROR(irot_rdq(MID_PRIORITY));
	SVC_PERROR(irot_rdq(LOW_PRIORITY));
}


/*
 *  output OMIKUJI
 *
 *  input : num : output pattern number (1...DEF_NUM_MAX)
 */
void  omikuji_print( int num )
{
    FRESULT  ret;
    FIL  fil;
    UINT  rdsz ;
    int  y ;
    int  i ;
    unsigned char  *pnt, mask ;
    char  namegra[32] ;
    char  *basegra = "01gra.bmp";

    print_area_clear();

    if( num <= 0 || num > var_data_num ) {
        num = 1 ;
    }

    ret = f_mount( 0, &var_fs2 );
    if( ret != FR_OK ) {
        syslog( LOG_NOTICE, "ERROR f_mount() = %d\n", ret );
    } else {
        syslog( LOG_NOTICE, "PASSED f_mount()" );
    }

    strcpy( namegra, basegra );
    namegra[0] = '0' + (num/10);
    namegra[1] = '0' + (num%10);
    syslog(LOG_NOTICE, "f_open(%s)", namegra );

    ret = f_open( &fil, namegra, FA_READ );
    if( ret != FR_OK ) {
        syslog( LOG_NOTICE, "ERROR f_open(%s) = %d\n", namegra, ret );
        while( 1 );
    }

    /* skip BMP file header */
    ret = f_read( &fil, readbuff, 0x3e, &rdsz );
    if( ret != FR_OK ) {
        syslog( LOG_NOTICE, "ERROR f_read(%s) = %d\n", namegra, ret );
    }

    for( i=0 ; i<384 ; i++ ) {
        /* get Y line */
        ret = f_read( &fil, readbuff, DEF_PRINT_BMP_A_Y, &rdsz );
        pnt = readbuff ;
        mask = 0x80 ;
        for( y=0 ; y<DEF_PRINT_AREA_Y ; y++ ) {
            if( !(*pnt & mask) ) print_pixel_set( i, y );
            mask >>= 1;
            if( mask == 0 ) {
                mask = 0x80 ;
                pnt++;
            }
        }
    }

    /* set bottom line */
    for( i=0 ; i<384 ; i++ ) {
        print_pixel_set( i, var_data_len*8-10 );
        print_pixel_set( i, var_data_len*8-11 );
    }

    print_output_area();
}


int  var_action_trig ;
int  var_cancel_trig ;

/*
 *  printout control task
 */
void  omikuji_print_task( intptr_t exinf )
{
    int  randnum, randcmp, patnum ;
    int  i ;
    int  debug_cnt ;

    syslog( LOG_NOTICE, "omikuji_print_task() start." );
    var_action_trig = 0;
    debug_cnt = 0 ;

    while( 1 ) {
        /* wait Action Trigger input */
        syslog( LOG_NOTICE, "wait Action Trigger." );
        while( var_action_trig == 0 ) {
	    /* 乱数列を進め、同じ並びにならないようにする */
            (void)rand();
            tslp_tsk(10);
        }

        /* make output pattern number */
        randnum = rand();
        randcmp = randnum % var_data_prb_total;
        for( i=0 ; i<var_data_num ; i++ ) {
            if( randcmp < var_data_prs[ i ]  ) {
                i++ ;
                break ;
            }
        }
        patnum  = i;
        if( var_debug_mode ) {
            patnum = (debug_cnt % var_data_num) + 1 ;
            debug_cnt++ ;
        }
        if( patnum <= 0 || patnum > var_data_num ) {
            patnum = 1 ;
        }

        syslog( LOG_NOTICE, "randnum=%d  randcmp=%d  patnum=%d", randnum, randcmp, patnum );
        /* 分布配列更新 */
        var_data_dst[ patnum-1 ]++ ;
        syslog( LOG_NOTICE, "[1]%d [2]%d [3]%d [4]%d [5]%d", 
                var_data_dst[0],var_data_dst[1],var_data_dst[2],var_data_dst[3],var_data_dst[4]);
        syslog( LOG_NOTICE, "[6]%d [7]%d [8]%d [9]%d [10]%d", 
                var_data_dst[5],var_data_dst[6],var_data_dst[7],var_data_dst[8],var_data_dst[9]);

        omikuji_print( patnum ) ;

        var_action_trig = 0;
    }
}


/*
 *  LED ON/OFF control task
 */
void  omikuji_led_task( intptr_t exinf )
{
    int  onoff_cnt, i ;
    int  onoff ;
    int  normal ;
    int  before ;

    syslog( LOG_NOTICE, "omikuji_led_task() start." );

    normal = 10 ;
    if( var_debug_mode ) {
       normal = 5 ;
    }

    onoff_cnt = normal ;
    onoff = 0 ;

    while( 1 ) {
	/* 乱数列を進め、同じ並びにならないようにする */
        (void)rand();

        before = var_action_trig ;
        if( before != 0 ) {
            onoff_cnt = 2 ;
        } else {
            onoff_cnt = normal ;
        }
        for( i=0 ; i<onoff_cnt ; i++ ) {
            tslp_tsk( 50 ) ;
        }

        onoff++ ;
        if( onoff & 1 ) {
            led_set( 0x01 );
        } else {
            led_set( 0x00 );
        }
    }
}


/*
 *  action trigger task
 */
void  omikuji_trigger_task( intptr_t exinf )
{
    int  cancel ;
    int  mask ;

    syslog( LOG_NOTICE, "omikuji_trigger_task() start." );

    var_cancel_trig = 0 ;
    cancel = 0 ;
    mask = 0x07 ;

    while( 1 ) {
        /* triggerの変化待ち */
        while( (btn_stat() & mask) == 0 ) {
            if( var_action_trig == 0 ) {
                var_cancel_trig = 0 ;
                /* 筒の読み出しを有効 */
                mask = 0x07 ;
            }
            tslp_tsk( 50 ) ;
        }

        if( var_action_trig == 1 ) {
            cancel = 1 ;
        }

        /* request Action */
        var_action_trig = 1 ;
        /* 筒の読み出しを無効 */
        mask = 0x05 ;

        /* triggerの変化が戻るのを待つ */
        while( btn_stat() != 0 ) {
            tslp_tsk( 50 ) ;
        }

        if( cancel == 1 && var_action_trig == 1 ) {
            var_cancel_trig = 1 ;
            cancel = 0 ;
        }
    }
}


/*
 *  OMIKUJI main sequence
 */
void  omikuji_main( void )
{
    DSTATUS  rc;
    FRESULT  ret;
    FIL  fil;
    UINT  rdsz ;
    int  i, j ;

    syslog( LOG_NOTICE, "SystemCoreClock = %d Hz", SystemCoreClock );
    uart_init();
    print_init();
    led_init();
    btn_init();

    var_debug_mode = 0 ;
    /* check onboard button */
    if( btn_stat() & 0x01) {
        /** set DEBUG MODE **/
        var_debug_mode = 1 ;
        syslog( LOG_NOTICE, "DEBUG MODE : ON" );
    }

    /* read data num */
    rc = disk_initialize( 0 );
    if( rc != RES_OK ) {
        syslog( LOG_NOTICE, "ERROR disk_initialize() = %d\n", rc );
    } else {
        syslog( LOG_NOTICE, "PASSED disk_initialize()" );
    }

    ret = f_mount( 0, &var_fs );
    if( ret != FR_OK ) {
        syslog( LOG_NOTICE, "ERROR f_mount() = %d\n", ret );
    } else {
        syslog( LOG_NOTICE, "PASSED f_mount()" );
    }

    var_data_num = 0 ;
    var_data_len = 0 ;

    /* get data num */
    ret = f_open( &fil, DEF_NUM_FILE, FA_READ );
    if( ret != FR_OK ) {
        syslog( LOG_NOTICE, "ERROR f_open(%s) = %d\n", DEF_NUM_FILE, ret );
    }
    syslog( LOG_NOTICE, "PASSED f_open(%s)", DEF_NUM_FILE );
    ret = f_read( &fil, readbuff, DEF_READBUFF, &rdsz );
    if( ret != FR_OK ) {
        syslog( LOG_NOTICE, "ERROR f_read(%s) = %d\n", DEF_NUM_FILE, ret );
    }
    f_close( &fil );
    syslog( LOG_NOTICE, "PASSED f_read(%s)", DEF_NUM_FILE );

    for( i=0 ; i<DEF_NUM_MAXCOL ; i++ ) {
        if( readbuff[i]>='0' && readbuff[i]<='9' ) {
            var_data_num *= 10 ;
            var_data_num += readbuff[i]-'0' ;
        } else {
            break ;
        }
    }
    if( var_data_num > DEF_NUM_MAX ) {
        var_data_num = DEF_NUM_MAX ;
    }
    syslog( LOG_NOTICE, "var_data_num = %d", var_data_num );

    /* get data Y len */
    ret = f_open( &fil, DEF_LEN_FILE, FA_READ );
    if( ret != FR_OK ) {
        syslog( LOG_NOTICE, "ERROR f_open(%s) = %d\n", DEF_LEN_FILE, ret );
    }
    syslog( LOG_NOTICE, "PASSED f_open(%s)", DEF_LEN_FILE );
    ret = f_read( &fil, readbuff, DEF_READBUFF, &rdsz );
    if( ret != FR_OK ) {
        syslog( LOG_NOTICE, "ERROR f_read(%s) = %d\n", DEF_LEN_FILE, ret );
    }
    f_close( &fil );
    syslog( LOG_NOTICE, "PASSED f_read(%s)", DEF_LEN_FILE );

    for( i=0 ; i<DEF_LEN_MAXCOL ; i++ ) {
        if( readbuff[i]>='0' && readbuff[i]<='9' ) {
            var_data_len *= 10 ;
            var_data_len += readbuff[i]-'0' ;
        } else {
            break ;
        }
    }
    if( var_data_len > DEF_LEN_MAX ) {
        var_data_len = DEF_LEN_MAX ;
    }
    syslog( LOG_NOTICE, "var_data_len = %d", var_data_len );

    /* get probability */
    ret = f_open( &fil, DEF_PRB_FILE, FA_READ );
    if( ret != FR_OK ) {
        syslog( LOG_NOTICE, "ERROR f_open(%s) = %d\n", DEF_PRB_FILE, ret );
    }
    syslog( LOG_NOTICE, "PASSED f_open(%s)", DEF_PRB_FILE );
    ret = f_read( &fil, readbuff, DEF_READBUFF, &rdsz );
    if( ret != FR_OK ) {
        syslog( LOG_NOTICE, "ERROR f_read(%s) = %d\n", DEF_PRB_FILE, ret );
    }
    f_close( &fil );
    syslog( LOG_NOTICE, "PASSED f_read(%s)", DEF_PRB_FILE );

    /* clear probability mat */
    for( i=0 ; i<DEF_NUM_MAX ; i++ ) {
        var_data_prb[ i ] = 0 ;
        var_data_prs[ i ] = 0 ;
        var_data_dst[ i ] = 0 ;
    }

    j = 0 ;
    for( i=0 ; i<rdsz ; i++ ) {
        if( readbuff[i]>='0' && readbuff[i]<='9' ) {
            var_data_prb[j] *= 10 ;
            var_data_prb[j] += readbuff[i]-'0' ;
        } else {
           for( ; i<rdsz ; i++ ) {
               if( readbuff[i] == 0x0a ) {
                   break ;
               }
           }
           j++ ;
        }
    }
    var_data_prb_total = 0;
    for( i=0 ; i<var_data_num ; i++ ) {
        var_data_prb_total += var_data_prb[i];
        var_data_prs[ i ] = var_data_prb_total;
        syslog( LOG_NOTICE, "var_data_prb[%2d] = %d", i, var_data_prb[i] );
    }
    syslog( LOG_NOTICE, "var_data_prb_total = %d", i, var_data_prb_total );

    /*
     *  タスクの起動
     */
    SVC_PERROR(act_tsk(TASK1));
    SVC_PERROR(act_tsk(TASK2));
    SVC_PERROR(act_tsk(TASK3));

    /* no return */
    while( 1 ) {
        tslp_tsk( 100 );
    }
}


/*
 *  メインタスク
 */
void main_task(intptr_t exinf)
{

	SVC_PERROR(syslog_msk_log(LOG_UPTO(LOG_INFO), LOG_UPTO(LOG_EMERG)));
	syslog(LOG_NOTICE, "Sample program starts (exinf = %d).", (int_t) exinf);

#if 0
    {
	ER_UINT	ercd;

	/*
	 *  シリアルポートの初期化
	 *
	 *  システムログタスクと同じシリアルポートを使う場合など，シリアル
	 *  ポートがオープン済みの場合にはここでE_OBJエラーになるが，支障は
	 *  ない．
	 */
	ercd = serial_opn_por(TASK_PORTID);
	if (ercd < 0 && MERCD(ercd) != E_OBJ) {
		syslog(LOG_ERROR, "%s (%d) reported by `serial_opn_por'.",
									itron_strerror(ercd), SERCD(ercd));
	}
	SVC_PERROR(serial_ctl_por(TASK_PORTID, (IOCTL_CRLF | IOCTL_FCSND | IOCTL_FCRCV)));
    }
#endif

	omikuji_main() ;
}
